"""
This file contains the client definition
"""

import socket
import serial
from typeguard import typechecked
from pyrsvp.rsvptypes import RsvpBaseType, process_data_types


class RsvpClientConnectionFailed(Exception):
    """Exception for client connection"""


class RsvpNotAck(Exception):
    """Exception when server doesn't acknowledge"""


class RsvpMalformedReply(Exception):
    """Reply is malformed"""


class RsvpTimeoutExpired(Exception):
    """Received or send timeout"""


class RsvpClientInvalidCallback(Exception):
    """Invalid callback function"""


class RsvpClient:
    """Base RsvpClient class"""

    def __init__(self):
        pass

    def _build_command(self, cmd_name: str, args: list[RsvpBaseType] = None) -> str:
        """Build the command string. The \r\n chars are also appended at the end.

        Params
        ------
        md_name : str
            Name of the command
        args : list
            List of arguments to append to the command

        Returns
        -------
        A string with the command
        """

        cmd_str = cmd_name + ";"

        if args:
            for arg in args:
                cmd_str = cmd_str + arg.__str__() + ";"

        return cmd_str[:-1] + "\r\n"

    def _check_response(self, cmd_name: str, resp: str) -> dict:
        """Check the server response and process it

        Params
        ------
        cmd_name : str
            Name of the command that generated this response
        resp : str
            Raw data received

        Raises
        ------
        RsvpClientNotAck: If something is wrong
        RsvpWrongTypeException: If it fails to parse the args
        """

        tokens = resp.split(";")

        cmd_reply = tokens[0]
        try:
            cmd, status = cmd_reply.split("=")
        except ValueError:
            raise RsvpMalformedReply(f"Malformed reply string: {resp}")

        if len(tokens) < 2:
            if status == "NACK":
                raise RsvpNotAck()

            return {}

        # You expect to have a REPLY-{cmd_name}
        if cmd_name in cmd:
            match status:
                case "NACK":
                    _, err_str = tokens[1].split("=")
                    raise RsvpNotAck(err_str)
                case "ACK":
                    tokens = tokens[1:]
                    return process_data_types(tokens)
                case _:
                    raise RsvpMalformedReply(f"Unknown reply {cmd_name}={status}")
        else:
            raise RsvpNotAck(
                f"Command {cmd_name} not included in reply: {cmd}. Server reply is malformed"
            )


class RsvpClientTcp(RsvpClient):
    """RsvpClient using TCP as transport layer"""

    def __init__(
        self,
        srv_ip_address: str = "localhost",
        srv_port: int = 8080,
        timeout: float = 5.0,
    ):
        """RsvpClientTcp constructor

        Params
        ------
        srv_ip_address : str
            IP address of the server
        srv_port : int
            Server port
        timeout : float
            Timeout used for communication with the server
        """

        self._srv_ip = srv_ip_address
        self._port = srv_port
        self._timeout = timeout

    def __connect(self):
        """Connect to the server"""

        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except Exception as ex:
            raise RsvpClientConnectionFailed(f"Failed to create socket: {ex}") from ex

        try:
            self.sock.settimeout(self._timeout)
        except Exception as ex:
            raise RsvpClientConnectionFailed(f"Failed to set timeout: {ex}") from ex

        try:
            self.sock.connect((self._srv_ip, self._port))
        except Exception as ex:
            raise RsvpClientConnectionFailed(f"Failed to connect: {ex}") from ex

    @typechecked
    def send_command(
        self, cmd_name: str, args: list[RsvpBaseType] = [], callback=None
    ) -> dict:
        """Send a command to the RSVP server and wait for a reply.

        Params
        ------
        cmd_name : str
            Name of the command
        args : list
            List of arguments to append to the command
        callback : function
            Function to call if the command is successful

        Returns
        -------
        The server response if no callback was given

        Raises
        ------
        RsvpTimeoutExpired : If while sending or receiving data the timeout expires
        RsvpClientInvalidCallback: If something is wrong with the callback
        RsvpMalformedReply: If the reply is not as expected
        RsvpNotAck: If the server didn't acknowledged
        """

        self.__connect()

        cmd = self._build_command(cmd_name, args)

        try:
            self.sock.sendall(cmd.encode())
        except socket.timeout as ex:
            raise RsvpTimeoutExpired(f"Failed to send data: {ex}") from ex

        try:
            buf = self.sock.recv(1024).strip()
        except socket.timeout as ex:
            raise RsvpTimeoutExpired(f"Failed to receive data: {ex}") from ex

        # Process the response
        data = self._check_response(cmd_name, buf.decode())
        self.sock.close()

        if callback:
            try:
                callback(data)
            except Exception as ex:
                raise RsvpClientInvalidCallback(ex) from ex

        return data


class RsvpClientSerial(RsvpClient):
    """RsvpClient using Serial as transport layer"""

    def __init__(self, port: str = "ttyUSB0", baud: int = 115200, timeout=1.0):
        """RsvpClientSerial constructor

        Params
        ------
        port : str
            Serial port name
        baud : int
            Baudrate of the serial port
        """
        super().__init__()
        self.ser = serial.Serial()
        self.ser.port = port
        self.ser.baudrate = baud
        self.ser.write_timeout = 1.0
        self.ser.timeout = timeout

        try:
            self.ser.open()
        except serial.SerialException as ex:
            raise RsvpClientConnectionFailed(ex.__str__()) from ex

    def __del__(self):
        self.ser.close()

    @typechecked
    def send_command(
        self, cmd_name: str, args: list[RsvpBaseType] = [], callback=None
    ) -> dict:
        """Send a command to the RSVP server and wait for a reply.

        Params
        ------
        cmd_name : str
            Name of the command
        args : list
            List of arguments to append to the command
        callback : function
            Function to call if the command is successful

        Returns
        -------
        The server response if no callback was given

        Raises
        ------
        RsvpTimeoutExpired : If while sending or receiving data the timeout expires
        RsvpNotAck: If the server didn't acknowledged
        RsvpClientInvalidCallback: If something is wrong with the callback
        RsvpMalformedReply: If the reply is not as expected
        """

        cmd = self._build_command(cmd_name, args)

        try:
            self.ser.write(cmd.encode())
        except serial.SerialTimeoutException as ex:
            raise RsvpTimeoutExpired(ex.__str__()) from ex

        data = self.ser.readline().strip()
        if len(data) < 1:
            raise RsvpTimeoutExpired(
                f"No data was received and the timeout expired. Try to change the timeout value to fix this issue"
            )

        data = self._check_response(cmd_name, data.decode())

        if callback:
            try:
                callback(data)
                return {}
            except Exception as ex:
                raise RsvpClientInvalidCallback(ex) from ex
        else:
            return data
