"""
This file contains the server definition
"""

import socketserver
import socket
import inspect
import serial
import signal
import time
from typeguard import typechecked

from pyrsvp.rsvptypes import process_data_types, RsvpWrongTypeException, RsvpBaseType


class RsvpNoExceptionCallback(Exception):
    """Exception if no exception callback defined"""


class RsvpUnknownException(Exception):
    """Unknown exception"""


class RsvpUnknownCommandException(Exception):
    """Unknown command"""


class RsvpServer:
    """RSVP server base class"""

    _command_dict: dict = {}
    _exception_cmd = "EXCEPTION"

    def _print_commands(self):
        """Prints the command list"""

        for key, value in self._command_dict.items():
            print(f"{key}: {value}")

    class Command:
        """Decorator to add commands"""

        def __init__(self, arg):
            self._arg = arg

        def __call__(self, *param_arg):
            RsvpServer._command_dict[self._arg] = param_arg[0]

    def _parse_data(self, data: str) -> str:
        """Parse the data received by the server and calls the correct
        command callback

        Params
        ------
        data : str
            Data to parse

        Returns
        -------
        An error message
        """

        # Split data
        tokens = data.split(";")

        cmd = tokens[0]

        # Remove command
        tokens = tokens[1:]

        try:
            values = process_data_types(tokens)
        except RsvpWrongTypeException as ex:
            if self._exception_cmd in self._command_dict:
                cmd = self._exception_cmd
                values = ex
            else:
                raise RsvpNoExceptionCallback(
                    'Implement a command named "EXCEPTION" to handle this case'
                ) from ex

        if cmd in self._command_dict:
            try:
                if inspect.ismethod(self._command_dict[cmd]):
                    reply = self._command_dict[cmd](values)
                else:
                    reply = self._command_dict[cmd](values)
            except Exception as ex:
                raise RsvpUnknownException(f"{type(ex).__name__}: {ex}") from ex
        else:
            if self._exception_cmd in self._command_dict:
                reply = self._command_dict[self._exception_cmd](
                    RsvpUnknownCommandException(f"{cmd}")
                )
            else:
                raise RsvpUnknownCommandException(f"{cmd}")

        return f"REPLY-{cmd}{reply}"

    @typechecked
    @staticmethod
    def reply(ack: bool, args: list[RsvpBaseType] = []) -> str:
        """Build a reply string

        Params
        ------
        ack : bool
            Acknowledge or not acknowledge
        args : list
            Args to append to the reply string

        Returns
        -------
        A reply string
        """

        reply = ""

        if ack:
            reply = "=ACK;"
        else:
            reply = "=NACK;"

        if args:
            for arg in args:
                reply = reply + arg.__str__() + ";"

        return reply[:-1] + "\r\n"


class RsvpServerTcp(RsvpServer):
    """RsvpServer over TCP"""

    def __init__(self, ip_address: str = "localhost", port: int = 8080):
        """Constructor

        Params
        ------
        ip_address : str
            IP address of the server
        port : int
            Port listened to
        """

        super().__init__()
        self.ip = ip_address
        self.port = port

    def print_commands(self):
        """Prints the command list"""
        return super()._print_commands()

    class _TcpHandler(socketserver.StreamRequestHandler, RsvpServer):
        """Request handler. One per client"""

        def setup(self) -> None:
            print(f"New client connected. IP address: {self.client_address[0]}")
            return super().setup()

        def handle(self) -> None:
            """Handle new data received from the client"""

            counter = 0

            while True:
                try:
                    data = self.rfile.readline().strip()
                except socket.timeout:
                    # Now we check if the socket is closed or if no data as been sent
                    try:
                        buf = self.server.socket.recv(
                            1, socket.MSG_PEEK | socket.MSG_DONTWAIT
                        )
                        if buf == b"":
                            continue
                    except Exception as ex:
                        print(ex)
                        break

                # If too many empty data is received consider that the client is dead
                if counter > 1000:
                    break

                if len(data) == 0:
                    counter = counter + 1
                    continue

                reply = super()._parse_data(data.decode())
                self.wfile.write(reply.encode())

        def finish(self) -> None:
            print(f"Client {self.client_address[0]} disconnected")
            return super().finish()

    def run(self):
        """Run the TCP server"""

        print(
            f"TCP server running. IP address: {self.ip}, port: {self.port}. Press CTRL+C to quit"
        )

        socketserver.TCPServer.allow_reuse_address = True

        with socketserver.TCPServer(
            (self.ip, self.port), RsvpServerTcp._TcpHandler
        ) as server:
            server.timeout = 1.0
            server.serve_forever()


class RsvpServerSerial(RsvpServer):
    """RsvpServer over serial"""

    def __init__(self, port: str = "ttyUSB0", baud: int = 115200):
        """RsvpServerSerial constructor

        Params
        ------
        port : str
            Port name
        baud : int
            Baudrate
        """

        super().__init__()
        self.ser = serial.Serial()
        self.ser.port = port
        self.ser.baudrate = baud
        self._running = False
        self.ser.timeout = 1
        signal.signal(signal.SIGINT, self.__sig_handler)

    def print_commands(self):
        """Prints the command list"""
        return super()._print_commands()

    def __sig_handler(self, signum, frame):
        if signum == signal.SIGINT:
            self._running = False
            self.ser.cancel_read()

    def run(self, debug=False):
        """Run the Serial server"""

        try:
            self.ser.open()
        except Exception as ex:
            raise ex

        self._running = True

        print(
            f"Serial server running. Com port: {self.ser.port}, baud: {self.ser.baudrate}. Press CTRL+C to quit"
        )

        while self._running:
            try:
                data = self.ser.readline()
            except serial.SerialException as ex:
                # Something went wrong probably an empty line was received
                print(ex)
                continue

            if len(data) == 0:
                continue

            if debug:
                print(f"[DEBUG] Data: {data}")

            data = data.strip()
            reply = super()._parse_data(data.decode())

            if debug:
                print(f"[DEBUG] Reply: {reply.encode()}")

            self.ser.write(reply.encode())

        self.ser.close()
