"""
This module contains the definition of the RSVP types
"""

from dataclasses import dataclass


class RsvpWrongTypeException(Exception):
    """Wrong type exception"""


@dataclass
class RsvpBaseType:
    """Base Rsvp type

    Raises
    ------
    RsvpWrongTypeException: If name is not a string
    """

    _name: str

    @property
    def name(self) -> str:
        """Variable name"""
        return self._name

    def __post_init__(self):
        if not isinstance(self._name, str):
            raise RsvpWrongTypeException("Name must be a string")


@dataclass
class RsvpInteger(RsvpBaseType):
    """Rsvp integer variable

    Raises
    ------
    RsvpWrongTypeException: If value in not an integer
    """

    _value: int

    @property
    def value(self) -> int:
        """Variable value"""
        return self._value

    def __post_init__(self):
        super().__post_init__()
        if not isinstance(self._value, int):
            raise RsvpWrongTypeException(f"Variable {self.name} is not an integer")

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(INT)=<value>
        """

        return f"{self._name}(INT)={self._value}"


@dataclass
class RsvpString(RsvpBaseType):
    """Rsvp string variable

    Raises
    ------
    RsvpWrongTypeException: If value in not a string
    """

    _value: str

    @property
    def value(self) -> str:
        """Variable value"""
        return self._value

    def __post_init__(self):
        super().__post_init__()
        if not isinstance(self._value, str):
            raise RsvpWrongTypeException(f"Variable {self._name} is not a string")

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(STR)=<value>
        """

        return f"{self._name}(STR)={self._value}"


@dataclass
class RsvpDouble(RsvpBaseType):
    """Rsvp double variable

    Raises
    ------
    RsvpWrongTypeException: If value in not a float
    """

    _value: float

    @property
    def value(self) -> float:
        """Variable value"""
        return self._value

    def __post_init__(self):
        super().__post_init__()
        if not isinstance(self._value, float):
            raise RsvpWrongTypeException(f"Variable {self._name} is not a float")

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(DBL)=<value>
        """

        return f"{self._name}(DBL)={self._value}"


@dataclass
class RsvpBool(RsvpBaseType):
    """Rsvp boolean variable

    Raises
    ------
    RsvpWrongTypeException: If value in not a boolean
    """

    _value: bool

    @property
    def value(self) -> bool:
        """Variable value"""
        return self._value

    def __post_init__(self):
        super().__post_init__()
        if not isinstance(self._value, bool):
            raise RsvpWrongTypeException(f"Variable {self._name} is not a boolean")

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(BOOL)=<value>
        """

        if self._value:
            _value = "TRUE"
        else:
            _value = "FALSE"

        return f"{self._name}(BOOL)={_value}"


@dataclass
class RsvpVectorDouble(RsvpBaseType):
    """Rsvp vector of double variable

    Raises
    ------
    RsvpWrongTypeException: If values in dict are not floats
    """

    _value: dict[str, float]

    @property
    def value(self) -> dict[str, float]:
        """Variable value"""
        return self._value

    def __post_init__(self):
        super().__post_init__()
        for key, val in self._value.items():
            if not isinstance(val, float):
                raise RsvpWrongTypeException(
                    f"In vector {self._name}. Variable {key} is not a double"
                )

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(VECD)=<value>
        """

        _str = f"{self._name}(VECD)=["

        for key, val in self._value.items():
            _str = _str + f"{key}={val},"

        _str = _str[:-1]

        return _str + "]"


@dataclass
class RsvpStringVector(RsvpBaseType):
    """Rsvp vector of double variable

    Raises
    ------
    RsvpWrongTypeException: If values in list are not strings
    """

    _value: list[str]

    @property
    def value(self) -> list[str]:
        """Variable value"""
        return self._value

    def __post_init__(self):
        for val in self._value:
            super().__post_init__()
            if not isinstance(val, str):
                raise RsvpWrongTypeException(
                    f"In vector {self._name}. Variable {val} is not a string"
                )

    def __str__(self) -> str:
        """Generate a RSVP string.

        Returns
        -------
        A string <name>(VECS)=<value>
        """

        _str = f"{self._name}(VECS)="

        for val in self._value:
            _str = _str + f"{val}\t"

        _str = _str[:-1]

        return _str


def process_data_types(tokens: list[str]) -> dict:
    """Process data from strings into a list with correct typing

    Params
    ------
    tokens : list[str]
        List with the data to process

    Raises
    ------
    RsvpWrongTypeException: If a data conversion fails
    """

    values = {}

    for token in tokens:
        # Ignore tokens with not enough characters
        if len(token) < 3:
            continue

        try:
            key, value = token.split("=", 1)
        except ValueError:
            # Ignore failed splits
            continue

        # Find type
        try:
            key_type = key[key.index("(") + 1 : key.index(")")]
        except ValueError:
            key_type = "STR"

        match key_type:
            case "INT":
                try:
                    value = int(value)
                except Exception as ex:
                    raise RsvpWrongTypeException(
                        f"{key}: {value} is not an integer"
                    ) from ex

            case "DBL":
                try:
                    value = float(value)
                except Exception as ex:
                    raise RsvpWrongTypeException(
                        f"{key}: {value} is not a double"
                    ) from ex

            case "STR":
                try:
                    value = str(value)
                except Exception as ex:
                    raise RsvpWrongTypeException(
                        f"{key}: {value} is not a string"
                    ) from ex

            case "BOOL":
                if value == "TRUE":
                    value = True
                elif value == "FALSE":
                    value = False
                else:
                    raise RsvpWrongTypeException(f"{key}: {value} is not a boolean")

            case "VECD":
                value = value[value.index("[") + 1 : value.index("]")]
                vals = value.split(",")

                _dict = {}

                for val in vals:
                    _key, _val = val.split("=")
                    try:
                        _dict[_key] = float(_val)
                    except Exception as ex:
                        raise RsvpWrongTypeException(
                            f"{key}: {value} is not a double"
                        ) from ex

                value = _dict

            case _:
                # No type found. Default is string
                try:
                    value = str(value)
                except Exception as ex:
                    raise RsvpWrongTypeException(
                        f"{key}: {value} is not a string"
                    ) from ex

        # Remove type from key
        try:
            key = key[: key.index("(")]
        except ValueError as ex:
            pass

        try:
            values[key] = value
        except Exception:
            pass

    return values
